result_type = NamedTuple{(:s, :e, :cpt, :CUSUM, :min_th, :scale),Tuple{Int64,Int64,Int64,Float64,Float64,Int64}}

function mad(data::Array)
    return  median(abs.(data .- median(data)))
end

#import StatsBase
#using StatsBase
# sigma = StatsBase.mad(x)
# result_type = @NamedTuple{s::Int64, e::Int64, cpt::Int64, CUSUM::Float64, min_th::Float64, scale::Int64}


"""
    WBS(data; th_const = 1.3, sigma = 0.0, M = 5000, do_seeded = false, shrink = 1/sqrt(2), min_length = 10)

Runs the Wild Binary segmentation algorithm for univariate array `data`, using the CUSUM cost function.
The test threshold is determined by `th_const` and the known or estimated standard deviation `sigma`. If `sigma = 0.0` is entered, this will be estimated by Median Absolute Deviance (MAD).
`do_seeded` determines whether to use seeded intervals. If true, `shrink` is the decay factor for interval length; if not, `M` random intervals are drawn, and returns the position of found changepoints, and
the cost of this segmentation. `min_length` determines minimum segment length.

See also: [`@BS`](@ref), [`@segment_cost`](@ref)

# Returns
* `result::Array{NamedTuple{s::Int64, e::Int64, cpt::Int64, CUSUM::Float64, min_th::Float64, scale::Int64}}`:
    * `s::Int64`: Start of window in which maximum CUSUM was located
    * `e::Int64`: End of window in which maximum CUSUM was located
    * `cpt::Int64`: Location of maximum CUSUM
    * `CUSUM::Float64`: Value of maximum CUSUM
    * `min_th::Float64`: Threshold used
    * `scale::Int64`: Number of recursion iterations before the change point was detected
# Example

```julia-repl
using Statistics

# Sample Normal time series with changing mean
n = 1000
λ = 100
μ, σ = Normal(0.0, 10.0), 1.0
sample, cps = @changepoint_sampler n λ Normal(μ, σ)
# Run wild binary segmentation
WBS_return = WBS(sample; sigma = σ)
```

# References
Fryzlewicz, P. (2014) Wild binary segmentation for multiple change-point detection, Annals of Statistics 42(6), 2243-2281
Kovács, S., Li, H., Bühlmann, P., & Munk, A. (2020). Seeded Binary Segmentation: A general methodology for fast and optimal change point detection. arXiv preprint arXiv:2002.06633.
"""
function WBS(data::Array; th_const=1.3, sigma=0.0, M=5000, do_seeded=false, shrink=1/sqrt(2), min_length=10)
    result = Array{result_type}(undef,0)

    if sigma == 0.0
        sigma = mad(data)
    end #if
    segment_cost = CUSUM(data, sigma)
    min_length= max(3, min_length)
    n = length(data)
    # threshold
    th = sigma * th_const * sqrt(2 * log(n))
    scale = 0
    # call WBS
    result = WBS_RECUR(segment_cost, 1, n, th, result, scale, M, do_seeded, shrink, min_length)

    return result
end


function WBS_RECUR(segment_cost::Function, s::Int64, e::Int64, th::Float64,
                   result::Array{result_type}, scale::Int64, M::Int64 = 5000, do_seeded::Bool = false, shrink::Float64 = 1/sqrt(2), min_length = 10)
    #k = length(CP) - 2 #current number of change points
     #minimum interval length
    scale += 1 #update scale

    # draw intervals
    if do_seeded
        tau = get_seeded_intervals(s,e, shrink, min_length)
        push!(tau, (s, e))
        M = length(tau)
    else
        tau = Tuple{Int, Int}[] # Segmentations to test
        push!(tau, (s, e))
        for m in 2:M
            s_m = rand(s:(e-min_length))
            e_m = rand( (s_m+min_length):e)
            push!(tau, (s_m, e_m))
        end
    end

    # compute CUSUMs find candidate change point
    x = fill(0.0,M, e-s+1)#zeros(Float64, M, e-s+1)#Array{Float64}(NaN, M, e-s+1 ) #CUSUM array
    for m in 1:M
        a, b = pop!(tau)
        for j in (a + 1):(b - 2)
            x[m,j-s+1] = segment_cost(s,e,j)
        end
    end
    #x = replace(x, 0.0 => maximum(x))
    maxval , ind = findmax(x)

    # segment here
    if maxval > th #&& scale < 10
        cpt = ind[2] + s -1
        newresult = (s=s,e=e, cpt=cpt, CUSUM = maxval, min_th=th, scale=scale )
        push!(result, newresult)
        #push!(CP,cpt )
        #push!(CUSUM, minval)
        if cpt - s > 2*min_length
            result = WBS_RECUR(segment_cost, s, cpt, th, result, scale, M, do_seeded, shrink, min_length)
        end
        if e - cpt > 2*min_length
            result = WBS_RECUR(segment_cost, cpt+1, e, th, result,  scale, M, do_seeded, shrink, min_length)
        end
    end
    #CP = sort(CP)

    return result #CUSUM, CP
end


"""
    get_WBS_changepoints(segment_cost, object, Kmax = 1,  alpha = 1.01)

Obtains change points from an `object` (output by the WBS function) by minimising an information criterion.
`segment_cost` is a cost function, normally sSIC, and `alpha` is the penalty exponent. `Kmax` determines the maximum number of changepoints to detect.

See also: [`WBS`](@ref), [`sSIC`](@ref), [`@segment_cost`](@ref)

# Returns
* `out_cps`: Vector of change point locations under optimal segmentation
* `optimal_number`: Integer number of change points which optimises the criterion
* `cost`: Vector of segmentation costs for 0, ..., `Kmax`  changes
* `object[1:optimal]`: WBS object truncated at optimal number
# Example

```julia-repl
# Sample Normal time series with changing mean
n = 1000
λ = 100
μ, σ = Normal(0.0, 10.0), 1.0
sample, cps = @changepoint_sampler n λ Normal(μ, σ)

# Run wild binary segmentation
WBS_return = WBS(sample)

# Obtain changepoints
seg_cost_sSIC = sSIC(sample)
get_WBS_changepoints(seg_cost_sSIC, WBS_return, 10)
```

# References
Fryzlewicz, P. (2014) Wild binary segmentation for multiple change-point detection, Annals of Statistics 42(6), 2243-2281
"""


function get_WBS_changepoints(segment_cost::Function, object::Array{result_type,1}, Kmax::Int64 = 1,  alpha::Float64 = 1.01)
    #result = Array{result_type}(undef,0)
    Kmax = min(Kmax,length(object))
    if length(object) > 0
        cost = zeros(Kmax+1)
        sorted = sort(object, lt = (a, b) -> ( a[4] < b[4]) ) #a[6] < b[6] &&
        cps_from_sorted = Vector{Int64}()
        for i in 1:Kmax
            push!(cps_from_sorted, sorted[i][3])
        end

        cost[1] = segment_cost(1, segment_cost.n)
        for i in 1:Kmax
            current_cps = sort(union([0, segment_cost.n], cps_from_sorted[1:i])) #select cps at current threshold
            for j in 1:(i+1)
                cost[i+1] = cost[i+1] + segment_cost(current_cps[j]+1, current_cps[j+1])
            end
        end

        pen = log(segment_cost.n)^alpha
        penvec = collect(0:Kmax) * pen
        cost += penvec

        optimal_number = findmin(cost)[2] -1

        out_cps = Vector{Int64}()
        for i in 1:optimal_number
            push!(out_cps, sorted[i][3])
        end
        out_cps = sort(out_cps)

        return  out_cps, optimal_number, cost, object[1:optimal_number]
    else
         println("no changepoints detected")
    end

end

function get_seeded_intervals(s::Int64, e::Int64,  shrink::Float64 = 1/sqrt(2), min_length = 10)
    n = e-s+1;
    max_k = Int(ceil(log(1/shrink, float(n) )))
    out = Tuple{Int, Int}[]

    for k = 1:max_k
        n_k = Int(2 * ceil(shrink^(1-k) ) - 1)
        l_k = n*shrink^(k-1)
        s_k = (n-l_k)/(n_k -1)
        if !isnan(s_k) && l_k >= min_length
            for i in 1:n_k
                left =  s + Int( floor((i-1)*s_k) )
                right = s + Int( ceil((i-1)*s_k + l_k) ) -1
                push!(out, (left, right) )
            end
        end
    end
    return out
end
